﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Collections;

namespace Achiral.Extension
{
    public static partial class MakeWorld
    {
        private partial class CovariantReadOnlyCollection<TSourceValue, TValue>
            : ICollection<TValue>
            where TSourceValue : TValue
        {
            private readonly ICollection<TSourceValue> _source;
            public CovariantReadOnlyCollection(ICollection<TSourceValue> source)
            {
                _source = source;
            }
            public void Add(TValue item)
            {
                throw new NotSupportedException();
            }
            public void Clear()
            {
                throw new NotSupportedException();
            }
            public bool Contains(TValue item)
            {
                return _source.Contains((TSourceValue)item);
            }
            public void CopyTo(TValue[] array, int arrayIndex)
            {
                if (array == null) throw new ArgumentNullException("array");
                if (arrayIndex < 0) throw new ArgumentOutOfRangeException();

                try
                {
                    checked
                    {
                        if (arrayIndex + this.Count > array.Length)
                        {
                            throw new ArgumentException("");
                        }
                        this.ForEach( (value, index) => array[arrayIndex + index] = value );
                    }
                }
                catch (OverflowException)
                {
                    throw new ArgumentException("");
                }
            }

            public int Count
            {
                get { return _source.Count; }
            }

            public bool IsReadOnly
            {
                get { return true; }
            }
            public bool Remove(TValue item)
            {
                throw new NotSupportedException();
            }
            public IEnumerator<TValue> GetEnumerator()
            {
                return _source.Select<TSourceValue, TValue>(value => value).GetEnumerator();
            }
            IEnumerator IEnumerable.GetEnumerator()
            {
                return _source.GetEnumerator();
            }
        }

        public static ICollection<TValue> AsReadOnly<TSourceValue, TValue>(this ICollection<TSourceValue> source)
            where TSourceValue : TValue
        {
            if (source == null) throw new ArgumentNullException("source");

            return new CovariantReadOnlyCollection<TSourceValue, TValue>(source);
        }
    }
}