﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Collections;

namespace Achiral.Extension
{
    public static partial class MakeWorld
    {
        private partial class CovariantReadOnlyDictionary<TSourceValue, TKey, TValue>
            : IDictionary<TKey, TValue>
            where TSourceValue : TValue
        {
            private readonly IDictionary<TKey, TSourceValue> _source;
            public CovariantReadOnlyDictionary(IDictionary<TKey, TSourceValue> source)
            {
                _source = source;
            }
            public void Add(TKey key, TValue value)
            {
                throw new NotSupportedException();
            }
            public bool ContainsKey(TKey key)
            {
                return _source.ContainsKey(key);
            }
            public ICollection<TKey> Keys
            {
                get { return _source.Keys; }
            }
            public bool Remove(TKey key)
            {
                throw new NotSupportedException();
            }
            public bool TryGetValue(TKey key, out TValue value)
            {
                TSourceValue soucevalue;
                bool result = _source.TryGetValue(key, out soucevalue);
                value = soucevalue;
                return result;
            }
            public ICollection<TValue> Values
            {
                get { return _source.Values.AsReadOnly<TSourceValue, TValue>(); }
            }
            public TValue this[TKey key]
            {
                get
                {
                    return _source[key];
                }
                set
                {
                    throw new NotSupportedException();
                }
            }
            public void Add(KeyValuePair<TKey, TValue> item)
            {
                throw new NotSupportedException();
            }
            public void Clear()
            {
                throw new NotSupportedException();
            }
            public bool Contains(KeyValuePair<TKey, TValue> item)
            {
                throw new NotImplementedException();
            }
            public void CopyTo(KeyValuePair<TKey, TValue>[] array, int arrayIndex)
            {
                if (array == null) throw new ArgumentNullException("array");
                if (arrayIndex < 0) throw new ArgumentOutOfRangeException();

                try
                {
                    checked
                    {
                        if (arrayIndex + this.Count > array.Length)
                        {
                            throw new ArgumentException("");
                        }
                        this.ForEach((value, index) => array[arrayIndex + index] = value);
                    }
                }
                catch (OverflowException)
                {
                    throw new ArgumentException("");
                }
            }
            public int Count
            {
                get { return _source.Count; }
            }
            public bool IsReadOnly
            {
                get { return true; }
            }
            public bool Remove(KeyValuePair<TKey, TValue> item)
            {
                throw new NotSupportedException();
            }
            public IEnumerator<KeyValuePair<TKey, TValue>> GetEnumerator()
            {
                foreach (var pair in _source)
                {
                    yield return new KeyValuePair<TKey, TValue>(pair.Key, pair.Value);
                }
            }
            IEnumerator IEnumerable.GetEnumerator()
            {
                return GetEnumerator();
            }
        }

        private partial class ReadOnlyLookup<TKey, TValue> : ILookup<TKey, TValue>
        {
            private readonly IDictionary<TKey, IEnumerable<TValue>> _dict;
            public ReadOnlyLookup(IDictionary<TKey, IEnumerable<TValue>> dictionary)
            {
                _dict = dictionary;
            }

            public bool Contains(TKey key)
            {
                return _dict.ContainsKey(key);
            }

            public int Count
            {
                get { return _dict.Count; }
            }

            public IEnumerable<TValue> this[TKey key]
            {
                get
                {
                    IEnumerable<TValue> enumerable;
                    if( _dict.TryGetValue(key, out enumerable))
                    {
                        return enumerable;
                    }
                    else
                    {
                        return Enumerable.Empty<TValue>();
                    }
                }
            }

            private class Grouping<TGKey, TGValue> : IGrouping<TGKey, TGValue>
            {
                private readonly ReadOnlyLookup<TGKey, TGValue> _lookup;
                private readonly TGKey _key;
                public Grouping(ReadOnlyLookup<TGKey, TGValue> lookup, TGKey key)
                {
                    this._key = key;
                    this._lookup = lookup;
                }

                public TGKey Key
                {
                    get { return _key; }
                }
                public IEnumerator<TGValue> GetEnumerator()
                {
                    return _lookup[_key].GetEnumerator();
                }
                IEnumerator IEnumerable.GetEnumerator()
                {
                    return GetEnumerator();
                }
            }

            public IEnumerator<IGrouping<TKey, TValue>> GetEnumerator()
            {
                foreach (var key in _dict.Keys)
                {
                    yield return new Grouping<TKey, TValue>(this, key);
                }
            }

            IEnumerator IEnumerable.GetEnumerator()
            {
                return GetEnumerator();
            }
        }

        public static IDictionary<TKey, TValue> AsReadOnly<TSourceValue, TKey, TValue>(this IDictionary<TKey, TSourceValue> source)
            where TSourceValue : TValue
        {
            if (source == null) throw new ArgumentNullException("source");

            return new CovariantReadOnlyDictionary<TSourceValue, TKey, TValue>(source);
        }

        public static ILookup<TKey, TValue> ToLookup<TKey, TValue>(this IDictionary<TKey, IEnumerable<TValue>> source)
        {
            if (source == null) throw new ArgumentNullException("source");

            return new ReadOnlyLookup<TKey, TValue>(source);
        }

    }


}